@@ -192,7 +192,7 @@ p11_mmap_open (const char *path,
 	if (map == NULL)
 		return NULL;
 
-	map->fd = open (path, O_RDONLY);
+	map->fd = open (path, O_RDONLY | O_CLOEXEC);
 	if (map->fd == -1) {
 		free (map);
 		return NULL;
@@ -298,14 +298,20 @@ p11_mmap_open (const char *path,
 	p11_mmap *map;
 
 	map = calloc (1, sizeof (p11_mmap));
-	if (map == NULL)
+	if (map == NULL) {
+		errno = ENOMEM;
 		return NULL;
+	}
 
 	map->file  = CreateFile (path, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, NULL);
 	if (map->file == INVALID_HANDLE_VALUE) {
 		errn = GetLastError ();
 		free (map);
 		SetLastError (errn);
+		if (errn == ERROR_PATH_NOT_FOUND || errn == ERROR_FILE_NOT_FOUND)
+			errno = ENOENT;
+		else if (errn == ERROR_ACCESS_DENIED)
+			errno = EPERM;
 		return NULL;
 	}
 
@@ -314,6 +320,8 @@ p11_mmap_open (const char *path,
 		CloseHandle (map->file);
 		free (map);
 		SetLastError (errn);
+		if (errn == ERROR_ACCESS_DENIED)
+			errno = EPERM;
 		return NULL;
 	}
 
@@ -323,6 +331,8 @@ p11_mmap_open (const char *path,
 		CloseHandle (map->file);
 		free (map);
 		SetLastError (errn);
+		if (errn == ERROR_ACCESS_DENIED)
+			errno = EPERM;
 		return NULL;
 	}
 
@@ -334,6 +344,8 @@ p11_mmap_open (const char *path,
 		CloseHandle (map->file);
 		free (map);
 		SetLastError (errn);
+		if (errn == ERROR_ACCESS_DENIED)
+			errno = EPERM;
 		return NULL;
 	}
 
@@ -676,7 +688,7 @@ _gettemp (char *path,
 
 	for (;;) {
 		if (doopen) {
-			if ((*doopen = open (path, O_BINARY | O_CREAT | O_EXCL | O_RDWR, 0600)) >= 0)
+			if ((*doopen = open (path, O_BINARY | O_CREAT | O_EXCL | O_RDWR | O_CLOEXEC, 0600)) >= 0)
 				return (1);
 			if (errno != EEXIST)
 				return (0);
@@ -747,3 +759,67 @@ mkdtemp (char *template)
 }
 
 #endif /* HAVE_MKDTEMP */
+
+#ifndef HAVE_GETAUXVAL
+
+unsigned long
+getauxval (unsigned long type)
+{
+	static unsigned long secure = 0UL;
+	static bool check_secure_initialized = false;
+
+	/*
+	 * This is the only one our stand-in impl supports and is
+	 * also the only type we define in compat.h header
+	 */
+	assert (type == AT_SECURE);
+
+	if (!check_secure_initialized) {
+#if defined(HAVE___LIBC_ENABLE_SECURE)
+		extern int __libc_enable_secure;
+		secure = __libc_enable_secure;
+
+#elif defined(HAVE_ISSETUGID)
+		secure = issetugid ();
+
+#elif defined(OS_UNIX)
+		uid_t ruid, euid, suid; /* Real, effective and saved user ID's */
+		gid_t rgid, egid, sgid; /* Real, effective and saved group ID's */
+
+#ifdef HAVE_GETRESUID
+		if (getresuid (&ruid, &euid, &suid) != 0 ||
+		    getresgid (&rgid, &egid, &sgid) != 0)
+#endif /* HAVE_GETRESUID */
+		{
+			suid = ruid = getuid ();
+			sgid = rgid = getgid ();
+			euid = geteuid ();
+			egid = getegid ();
+		}
+
+		secure = (ruid != euid || ruid != suid ||
+		          rgid != egid || rgid != sgid);
+#endif /* OS_UNIX */
+		check_secure_initialized = true;
+	}
+
+	return secure;
+}
+
+#endif /* HAVE_GETAUXVAL */
+
+#ifndef HAVE_STRERROR_R
+
+int
+strerror_r (int errnum,
+            char *buf,
+            size_t buflen)
+{
+#ifdef OS_WIN32
+	return strerror_s (buf, buflen, errnum);
+#else
+	#error no strerror_r implementation
+#endif
+}
+
+#endif /* HAVE_STRERROR_R */
