@@ -39,6 +39,10 @@
 
 #include <sys/types.h>
 
+#ifdef _GNU_SOURCE
+#error Make the crap stop. _GNU_SOURCE is completely unportable and breaks all sorts of behavior
+#endif
+
 #if !defined(__cplusplus) && (__GNUC__ > 2)
 #define GNUC_PRINTF(x, y) __attribute__((__format__(__printf__, x, y)))
 #else
@@ -68,6 +72,10 @@
 #define O_BINARY 0
 #endif
 
+#ifndef O_CLOEXEC
+#define O_CLOEXEC 0
+#endif
+
 #ifndef HAVE_GETPROGNAME
 const char * getprogname (void);
 #endif
@@ -103,6 +111,8 @@ char *       strdup_path_mangle (const char *template);
 #define WIN32_LEAN_AND_MEAN 1
 #include <windows.h>
 
+#include <io.h>
+
 /* Oh ... my ... god */
 #undef CreateMutex
 
@@ -164,6 +174,7 @@ void        p11_mmap_close  (p11_mmap *map);
 #include <pthread.h>
 #include <dlfcn.h>
 #include <time.h>
+#include <unistd.h>
 
 typedef pthread_mutex_t p11_mutex_t;
 
@@ -259,7 +270,7 @@ char *     strconcat        (const char *first,
 
 #endif /* HAVE_STRCONCAT */
 
-#ifndef HAVE_ASPRINTF
+#if defined HAVE_DECL_ASPRINTF && !HAVE_DECL_ASPRINTF
 
 int        asprintf         (char **strp,
                              const char *fmt,
@@ -267,14 +278,14 @@ int        asprintf         (char **strp,
 
 #endif /* HAVE_ASPRINTF */
 
-#ifndef HAVE_VASPRINTF
+#if defined HAVE_DECL_VASPRINTF && !HAVE_DECL_VASPRINTF
 #include <stdarg.h>
 
 int        vasprintf        (char **strp,
                              const char *fmt,
                              va_list ap);
 
-#endif /* HAVE_VASPRINTF */
+#endif /* HAVE_DECL_VASPRINTF */
 
 #ifndef HAVE_GMTIME_R
 #include <time.h>
@@ -291,4 +302,24 @@ time_t      timegm          (struct tm *tm);
 
 #endif /* HAVE_TIMEGM */
 
+#ifdef HAVE_GETAUXVAL
+
+#include <sys/auxv.h>
+
+#else /* !HAVE_GETAUXVAL */
+
+unsigned long     getauxval (unsigned long type);
+
+#define AT_SECURE 23
+
+#endif /* !HAVE_GETAUXVAL */
+
+#ifndef HAVE_STRERROR_R
+
+int         strerror_r      (int errnum,
+                             char *buf,
+                             size_t buflen);
+
+#endif /* HAVE_STRERROR_R */
+
 #endif /* __COMPAT_H__ */
