@@ -36,16 +36,25 @@
 
 #define P11_TEST_SOURCE 1
 
+#include "compat.h"
 #include "test.h"
 #include "debug.h"
+#include "path.h"
 
 #include <assert.h>
+#include <errno.h>
 #include <setjmp.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
+#ifdef OS_UNIX
+#include <sys/stat.h>
+#include <sys/wait.h>
+#include <unistd.h>
+#endif
+
 enum {
 	FIXTURE,
 	TEST,
@@ -121,6 +130,13 @@ p11_test_fail (const char *filename,
 	printf ("# in %s() at %s:%d\n", function, filename, line);
 
 	free (output);
+
+	/* Let coverity know we're not supposed to return from here */
+#ifdef __COVERITY__
+	abort();
+#endif
+
+	longjmp (gl.jump, 1);
 }
 
 static void
@@ -198,6 +214,7 @@ p11_test_run (int argc,
 	test_item *next;
 	int count;
 	int ret = 0;
+	int setup;
 
 	/* p11-kit specific stuff */
 	putenv ("P11_KIT_STRICT=1");
@@ -227,20 +244,27 @@ p11_test_run (int argc,
 		assert (item->type == TEST);
 		gl.last = item;
 		gl.number++;
+		setup = 0;
 
 		if (setjmp (gl.jump) == 0) {
 			if (fixture && fixture->x.fix.setup)
 				(fixture->x.fix.setup) (item->x.test.argument);
 
+			setup = 1;
+
 			assert (item->x.test.func);
 			(item->x.test.func)(item->x.test.argument);
 
-			if (fixture && fixture->x.fix.teardown)
-				(fixture->x.fix.teardown) (item->x.test.argument);
-
 			printf ("ok %d %s\n", gl.number, item->x.test.name);
 		}
 
+		if (setup) {
+			if (setjmp (gl.jump) == 0) {
+				if (fixture && fixture->x.fix.teardown)
+					(fixture->x.fix.teardown) (item->x.test.argument);
+			}
+		}
+
 		gl.last = NULL;
 	}
 
@@ -259,3 +283,151 @@ p11_test_run (int argc,
 	gl.number = 0;
 	return ret;
 }
+
+static char *
+expand_tempdir (const char *name)
+{
+	const char *env;
+
+	env = getenv ("TMPDIR");
+	if (env && env[0]) {
+		return p11_path_build (env, name, NULL);
+
+	} else {
+#ifdef OS_UNIX
+#ifdef _PATH_TMP
+		return p11_path_build (_PATH_TMP, name, NULL);
+#else
+		return p11_path_build ("/tmp", name, NULL);
+#endif
+
+#else /* OS_WIN32 */
+		char directory[MAX_PATH + 1];
+
+		if (!GetTempPathA (MAX_PATH + 1, directory)) {
+			printf ("# couldn't lookup temp directory\n");
+			errno = ENOTDIR;
+			return NULL;
+		}
+
+		return p11_path_build (directory, name, NULL);
+
+#endif /* OS_WIN32 */
+	}
+}
+
+char *
+p11_test_directory (const char *prefix)
+{
+	char *templ;
+	char *directory;
+
+	if (asprintf (&templ, "%s.XXXXXX", prefix) < 0)
+		assert_not_reached ();
+
+	directory = expand_tempdir (templ);
+	assert (directory != NULL);
+
+	if (!mkdtemp (directory)) {
+		printf ("# couldn't create temp directory: %s: %s\n",
+		        directory, strerror (errno));
+		free (directory);
+		assert_not_reached ();
+	}
+
+	free (templ);
+	return directory;
+}
+
+#ifdef OS_UNIX
+
+static void
+copy_file (const char *input,
+           int fd)
+{
+	p11_mmap *mmap;
+	const char *data;
+	ssize_t written;
+	size_t size;
+
+	mmap = p11_mmap_open (input, (void **)&data, &size);
+	assert (mmap != NULL);
+
+	while (size > 0) {
+		written = write (fd, data, size);
+		assert (written >= 0);
+
+		data += written;
+		size -= written;
+	}
+
+	p11_mmap_close (mmap);
+}
+
+char *
+p11_test_copy_setgid (const char *input)
+{
+	gid_t groups[128];
+		char *path;
+		gid_t group = 0;
+		int ret;
+		int fd;
+		int i;
+
+	ret = getgroups (128, groups);
+	for (i = 0; i < ret; ++i) {
+		if (groups[i] != getgid ()) {
+			group = groups[i];
+			break;
+		}
+	}
+	if (i == ret) {
+		fprintf (stderr, "# no suitable group, skipping test\n");
+		return NULL;
+	}
+
+	path = strdup ("/tmp/test-setgid.XXXXXX");
+	assert (path != NULL);
+
+	fd = mkstemp (path);
+	assert (fd >= 0);
+
+	copy_file (input, fd);
+	if (fchown (fd, getuid (), group) < 0)
+		assert_not_reached ();
+	if (fchmod (fd, 02750) < 0)
+		assert_not_reached ();
+	if (close (fd) < 0)
+		assert_not_reached ();
+
+	return path;
+}
+
+int
+p11_test_run_child (const char **argv,
+                    bool quiet_out)
+{
+	pid_t child;
+	int status;
+
+	child = fork ();
+	assert (child >= 0);
+
+	/* In the child process? */
+	if (child == 0) {
+		if (quiet_out)
+			close (1); /* stdout */
+		execve (argv[0], (char **)argv, NULL);
+		assert_not_reached ();
+	}
+
+	if (waitpid (child, &status, 0) < 0)
+		assert_not_reached ();
+
+	assert (!WIFSIGNALED (status));
+	assert (WIFEXITED (status));
+
+	return WEXITSTATUS (status);
+}
+
+#endif /* OS_UNIX */
