@@ -118,8 +118,7 @@ _p11_conf_parse_file (const char* filename, int flags)
 			p11_debug ("config file is inaccessible");
 
 		} else {
-			p11_message ("couldn't open config file: %s: %s", filename,
-			             strerror (error));
+			p11_message_err (error, "couldn't open config file: %s", filename);
 			errno = error;
 			return NULL;
 		}
@@ -288,8 +287,8 @@ static char *
 calc_name_from_filename (const char *fname)
 {
 	/* We eventually want to settle on .module */
-	static const char *suffix = ".module";
-	static size_t suffix_len = 7;
+	static const char *const suffix = ".module";
+	static const size_t suffix_len = 7;
 	const char *c = fname;
 	size_t fname_len;
 	size_t name_len;
@@ -400,13 +399,11 @@ load_configs_from_directory (const char *directory,
 			p11_debug ("couldn't list inacessible module configs");
 			return true;
 		}
-		p11_message ("couldn't list directory: %s: %s", directory,
-		             strerror (error));
+		p11_message_err (error, "couldn't list directory: %s", directory);
 		errno = error;
 		return false;
 	}
 
-	/* We're within a global mutex, so readdir is safe */
 	while ((dp = readdir(dir)) != NULL) {
 		path = p11_path_build (directory, dp->d_name, NULL);
 		return_val_if_fail (path != NULL, false);
@@ -419,7 +416,7 @@ load_configs_from_directory (const char *directory,
 		{
 			if (stat (path, &st) < 0) {
 				error = errno;
-				p11_message ("couldn't stat path: %s", path);
+				p11_message_err (error, "couldn't stat path: %s", path);
 				free (path);
 				break;
 			}
