@@ -53,7 +53,7 @@ struct {
 static void
 setup (void *unused)
 {
-	test.index = p11_index_new (NULL, NULL, NULL);
+	test.index = p11_index_new (NULL, NULL, NULL, NULL);
 	assert_ptr_not_null (test.index);
 }
 
@@ -397,6 +397,7 @@ handles_are (CK_OBJECT_HANDLE *handles,
              ...)
 {
 	CK_OBJECT_HANDLE handle;
+	bool matched = true;
 	int count;
 	int num;
 	va_list va;
@@ -410,7 +411,7 @@ handles_are (CK_OBJECT_HANDLE *handles,
 
 	va_start (va, handles);
 
-	for (count = 0; true; count++) {
+	for (count = 0; matched; count++) {
 		handle = va_arg (va, CK_OBJECT_HANDLE);
 		if (handle == 0)
 			break;
@@ -421,12 +422,12 @@ handles_are (CK_OBJECT_HANDLE *handles,
 		}
 
 		if (handles[i] != handle)
-			return false;
+			matched = false;
 	}
 
 	va_end (va);
 
-	return (count == num);
+	return matched && (count == num);
 }
 
 static void
@@ -653,26 +654,72 @@ test_replace_all (void)
 	assert_num_eq (4, p11_index_size (test.index));
 }
 
+static CK_RV
+on_index_build_fail (void *data,
+                     p11_index *index,
+                     CK_ATTRIBUTE *attrs,
+                     CK_ATTRIBUTE *merge,
+                     CK_ATTRIBUTE **populate)
+{
+	CK_ATTRIBUTE *match = data;
+
+	if (p11_attrs_match (merge, match))
+		return CKR_FUNCTION_FAILED;
+
+	return CKR_OK;
+}
+
+static void
+test_replace_all_build_fails (void)
+{
+	CK_ATTRIBUTE replace[] = {
+		{ CKA_LABEL, "odd", 3 },
+		{ CKA_VALUE, "one", 3 },
+		{ CKA_APPLICATION, "test", 4 },
+		{ CKA_INVALID }
+	};
+
+	CK_ATTRIBUTE match[] = {
+		{ CKA_LABEL, "odd", 3 },
+		{ CKA_INVALID }
+	};
+
+	p11_array *array;
+	p11_index *index;
+	CK_RV rv;
+
+	index = p11_index_new (on_index_build_fail, NULL, NULL, &match);
+	assert_ptr_not_null (index);
+
+	array = p11_array_new (p11_attrs_free);
+	if (!p11_array_push (array, p11_attrs_dup (replace)))
+		assert_not_reached ();
+
+	rv = p11_index_replace_all (index, NULL, CKA_INVALID, array);
+	assert_num_eq (rv, CKR_FUNCTION_FAILED);
+
+	p11_array_free (array);
+	p11_index_free (index);
+}
+
 
 static CK_RV
 on_build_populate (void *data,
                    p11_index *index,
-                   CK_ATTRIBUTE **attrs,
-                   CK_ATTRIBUTE *merge)
+                   CK_ATTRIBUTE *attrs,
+                   CK_ATTRIBUTE *merge,
+                   CK_ATTRIBUTE **populate)
 {
-	CK_ATTRIBUTE override[] = {
+	CK_ATTRIBUTE more[] = {
 		{ CKA_APPLICATION, "vigorous", 8 },
 		{ CKA_LABEL, "naay", 4 },
-		{ CKA_INVALID },
 	};
 
 	assert_str_eq (data, "blah");
 	assert_ptr_not_null (index);
-	assert_ptr_not_null (attrs);
 	assert_ptr_not_null (merge);
 
-	*attrs = p11_attrs_merge (*attrs, merge, true);
-	*attrs = p11_attrs_merge (*attrs, p11_attrs_dup (override), true);
+	*populate = p11_attrs_buildn (*populate, more, 2);
 	return CKR_OK;
 }
 
@@ -687,7 +734,7 @@ test_build_populate (void)
 	};
 
 	CK_ATTRIBUTE after[] = {
-		{ CKA_LABEL, "naay", 4 },
+		{ CKA_LABEL, "yay", 3 },
 		{ CKA_VALUE, "eight", 5 },
 		{ CKA_APPLICATION, "vigorous", 8 },
 		{ CKA_INVALID }
@@ -698,7 +745,7 @@ test_build_populate (void)
 	p11_index *index;
 	CK_RV rv;
 
-	index = p11_index_new (on_build_populate, NULL, "blah");
+	index = p11_index_new (on_build_populate, NULL, NULL, "blah");
 	assert_ptr_not_null (index);
 
 	rv = p11_index_add (index, original, 2, &handle);
@@ -723,8 +770,9 @@ test_build_populate (void)
 static CK_RV
 on_build_fail (void *data,
                p11_index *index,
-               CK_ATTRIBUTE **attrs,
-               CK_ATTRIBUTE *merge)
+               CK_ATTRIBUTE *attrs,
+               CK_ATTRIBUTE *merge,
+               CK_ATTRIBUTE **populate)
 {
 	CK_ATTRIBUTE check[] = {
 		{ CKA_LABEL, "nay", 3 },
@@ -737,7 +785,6 @@ on_build_fail (void *data,
 	if (p11_attrs_match (merge, check))
 		return CKR_DEVICE_ERROR;
 
-	*attrs = p11_attrs_merge (*attrs, merge, true);
 	return CKR_OK;
 }
 
@@ -761,7 +808,7 @@ test_build_fail (void)
 	p11_index *index;
 	CK_RV rv;
 
-	index = p11_index_new (on_build_fail, NULL, "testo");
+	index = p11_index_new (on_build_fail, NULL, NULL, "testo");
 	assert_ptr_not_null (index);
 
 	rv = p11_index_add (index, okay, 2, &handle);
@@ -825,7 +872,7 @@ test_change_called (void)
 	p11_index *index;
 	CK_RV rv;
 
-	index = p11_index_new (NULL, on_change_check, "change-check");
+	index = p11_index_new (NULL, NULL, on_change_check, "change-check");
 	assert_ptr_not_null (index);
 
 	on_change_removing = false;
@@ -870,15 +917,15 @@ test_change_batch (void)
 	p11_index *index;
 	CK_RV rv;
 
-	index = p11_index_new (NULL, on_change_check, "change-check");
+	index = p11_index_new (NULL, NULL, on_change_check, "change-check");
 	assert_ptr_not_null (index);
 
 	on_change_batching = true;
 	on_change_called = 0;
 
-	p11_index_batch (index);
+	p11_index_load (index);
 
-	assert (p11_index_in_batch (index));
+	assert (p11_index_loading (index));
 
 	rv = p11_index_add (index, original, 2, NULL);
 	assert (rv == CKR_OK);
@@ -896,7 +943,7 @@ test_change_batch (void)
 	assert_num_eq (0, on_change_called);
 
 	/* Nested batch is a noop */
-	p11_index_batch (index);
+	p11_index_load (index);
 
 	rv = p11_index_remove (index, handle);
 	assert (rv == CKR_OK);
@@ -909,7 +956,7 @@ test_change_batch (void)
 	 */
 	p11_index_finish (index);
 
-	assert (!p11_index_in_batch (index));
+	assert (!p11_index_loading (index));
 
 	/*
 	 * Only three calls, because later operations on the
@@ -920,7 +967,7 @@ test_change_batch (void)
 	/* This is a noop */
 	p11_index_finish (index);
 
-	assert (!p11_index_in_batch (index));
+	assert (!p11_index_loading (index));
 
 	p11_index_free (index);
 }
@@ -961,7 +1008,7 @@ test_change_nested (void)
 	p11_index *index;
 	CK_RV rv;
 
-	index = p11_index_new (NULL, on_change_nested, "change-nested");
+	index = p11_index_new (NULL, NULL, on_change_nested, "change-nested");
 	assert_ptr_not_null (index);
 
 	on_change_called = 0;
@@ -971,7 +1018,7 @@ test_change_nested (void)
 
 
 	on_change_called = 0;
-	p11_index_batch (index);
+	p11_index_load (index);
 	rv = p11_index_add (index, original, 2, NULL);
 	assert (rv == CKR_OK);
 	p11_index_finish (index);
@@ -1006,5 +1053,7 @@ main (int argc,
 	p11_test (test_change_called, "/index/change_called");
 	p11_test (test_change_batch, "/index/change_batch");
 	p11_test (test_change_nested, "/index/change_nested");
+	p11_test (test_replace_all_build_fails, "/index/replace-all-build-fails");
+
 	return p11_test_run (argc, argv);
 }
