@@ -43,9 +43,10 @@
 #include <string.h>
 
 #include "attrs.h"
-#include "hash.h"
+#include "digest.h"
 #include "library.h"
 #include "path.h"
+#include "parser.h"
 #include "pkcs11x.h"
 #include "token.h"
 
@@ -60,10 +61,15 @@
 #define NUM_SLOTS 3
 
 static CK_OBJECT_CLASS data = CKO_DATA;
+static CK_BBOOL vtrue = CK_TRUE;
+static CK_BBOOL vfalse = CK_FALSE;
 
 struct {
 	CK_FUNCTION_LIST *module;
 	CK_SLOT_ID slots[NUM_SLOTS];
+	char *directory;
+	p11_asn1_cache *cache;
+	p11_parser *parser;
 } test;
 
 static void
@@ -106,12 +112,55 @@ teardown (void *unused)
 {
 	CK_RV rv;
 
+	if (test.parser)
+		p11_parser_free (test.parser);
+	p11_asn1_cache_free (test.cache);
+
 	rv = test.module->C_Finalize (NULL);
 	assert (rv == CKR_OK);
 
+	free (test.directory);
+
 	memset (&test, 0, sizeof (test));
 }
 
+static void
+setup_writable (void *unused)
+{
+	CK_C_INITIALIZE_ARGS args;
+	char *arguments;
+	CK_ULONG count;
+	CK_RV rv;
+
+	memset (&test, 0, sizeof (test));
+
+	/* This is the entry point of the trust module, linked to this test */
+	rv = C_GetFunctionList (&test.module);
+	assert (rv == CKR_OK);
+
+	test.directory = p11_test_directory ("test-module");
+
+	memset (&args, 0, sizeof (args));
+	if (asprintf (&arguments, "paths='%s'", test.directory) < 0)
+		assert (false && "not reached");
+	args.pReserved = arguments;
+	args.flags = CKF_OS_LOCKING_OK;
+
+	rv = test.module->C_Initialize (&args);
+	assert (rv == CKR_OK);
+
+	free (arguments);
+
+	count = 1;
+	rv = test.module->C_GetSlotList (CK_TRUE, test.slots, &count);
+	assert_num_eq (rv, CKR_OK);
+	assert_num_eq (count, 1);
+
+	test.cache = p11_asn1_cache_new ();
+	test.parser = p11_parser_new (test.cache);
+	p11_parser_formats (test.parser, p11_parser_format_persist, NULL);
+}
+
 static void
 test_get_slot_list (void)
 {
@@ -139,6 +188,71 @@ test_get_slot_list (void)
 		assert (slots[i] != 0);
 }
 
+static void
+test_null_initialize (void)
+{
+	CK_FUNCTION_LIST *module;
+	CK_RV rv;
+
+	/* This is the entry point of the trust module, linked to this test */
+	rv = C_GetFunctionList (&module);
+	assert_num_eq (rv, CKR_OK);
+
+	rv = module->C_Initialize (NULL);
+	assert_num_eq (rv, CKR_OK);
+
+	rv = module->C_Finalize (NULL);
+	assert_num_eq (CKR_OK, rv);
+}
+
+static void
+test_multi_initialize (void)
+{
+	static CK_C_INITIALIZE_ARGS args =
+		{ NULL, NULL, NULL, NULL, CKF_OS_LOCKING_OK, NULL, };
+	CK_FUNCTION_LIST *module;
+	CK_SESSION_HANDLE session;
+	CK_SLOT_ID slots[8];
+	CK_SESSION_INFO info;
+	CK_ULONG count;
+	CK_RV rv;
+
+	/* This is the entry point of the trust module, linked to this test */
+	rv = C_GetFunctionList (&module);
+	assert_num_eq (rv, CKR_OK);
+
+	rv = module->C_Initialize (&args);
+	assert_num_eq (rv, CKR_OK);
+
+	count = 8;
+	rv = module->C_GetSlotList (CK_TRUE, slots, &count);
+	assert_num_eq (rv, CKR_OK);
+	assert_num_cmp (count, >, 0);
+
+	rv = module->C_OpenSession (slots[0], CKF_SERIAL_SESSION, NULL, NULL, &session);
+	assert_num_eq (rv, CKR_OK);
+
+	rv = module->C_GetSessionInfo (session, &info);
+	assert_num_eq (rv, CKR_OK);
+	assert_num_eq (info.slotID, slots[0]);
+
+	rv = module->C_Initialize (&args);
+	assert_num_eq (rv, CKR_OK);
+
+	rv = module->C_GetSessionInfo (session, &info);
+	assert_num_eq (rv, CKR_OK);
+	assert_num_eq (info.slotID, slots[0]);
+
+	rv = module->C_Finalize (NULL);
+	assert_num_eq (CKR_OK, rv);
+
+	rv = module->C_Finalize (NULL);
+	assert_num_eq (CKR_OK, rv);
+
+	rv = module->C_Finalize (NULL);
+	assert_num_eq (CKR_CRYPTOKI_NOT_INITIALIZED, rv);
+}
+
 static void
 test_get_slot_info (void)
 {
@@ -344,7 +458,6 @@ check_trust_object_equiv (CK_SESSION_HANDLE session,
 	unsigned char subject[1024];
 	unsigned char issuer[1024];
 	unsigned char serial[128];
-	CK_BBOOL modifiable;
 	CK_BBOOL private;
 	CK_BBOOL token;
 	CK_RV rv;
@@ -353,14 +466,13 @@ check_trust_object_equiv (CK_SESSION_HANDLE session,
 	CK_ATTRIBUTE equiv[] = {
 		{ CKA_TOKEN, &token, sizeof (token) },
 		{ CKA_PRIVATE, &private, sizeof (private) },
-		{ CKA_MODIFIABLE, &modifiable, sizeof (modifiable) },
 		{ CKA_ISSUER, issuer, sizeof (issuer) },
 		{ CKA_SUBJECT, subject, sizeof (subject) },
 		{ CKA_SERIAL_NUMBER, serial, sizeof (serial) },
 		{ CKA_INVALID, },
 	};
 
-	rv = test.module->C_GetAttributeValue (session, trust, equiv, 6);
+	rv = test.module->C_GetAttributeValue (session, trust, equiv, 5);
 	assert_num_eq (CKR_OK, rv);
 
 	test_check_attrs (equiv, cert);
@@ -371,8 +483,8 @@ check_trust_object_hashes (CK_SESSION_HANDLE session,
                            CK_OBJECT_HANDLE trust,
                            CK_ATTRIBUTE *cert)
 {
-	unsigned char sha1[P11_HASH_SHA1_LEN];
-	unsigned char md5[P11_HASH_MD5_LEN];
+	unsigned char sha1[P11_DIGEST_SHA1_LEN];
+	unsigned char md5[P11_DIGEST_MD5_LEN];
 	unsigned char check[128];
 	CK_ATTRIBUTE *value;
 	CK_RV rv;
@@ -389,10 +501,10 @@ check_trust_object_hashes (CK_SESSION_HANDLE session,
 	value = p11_attrs_find_valid (cert, CKA_VALUE);
 	assert_ptr_not_null (value);
 
-	p11_hash_md5 (check, value->pValue, value->ulValueLen, NULL);
+	p11_digest_md5 (check, value->pValue, value->ulValueLen, NULL);
 	assert (memcmp (md5, check, sizeof (md5)) == 0);
 
-	p11_hash_sha1 (check, value->pValue, value->ulValueLen, NULL);
+	p11_digest_sha1 (check, value->pValue, value->ulValueLen, NULL);
 	assert (memcmp (sha1, check, sizeof (sha1)) == 0);
 }
 
@@ -436,7 +548,6 @@ check_certificate (CK_SESSION_HANDLE session,
 	CK_DATE start;
 	CK_DATE end;
 	CK_ULONG category;
-	CK_BBOOL modifiable;
 	CK_BBOOL private;
 	CK_BBOOL token;
 	CK_RV rv;
@@ -445,7 +556,6 @@ check_certificate (CK_SESSION_HANDLE session,
 		{ CKA_CLASS, &klass, sizeof (klass) },
 		{ CKA_TOKEN, &token, sizeof (token) },
 		{ CKA_PRIVATE, &private, sizeof (private) },
-		{ CKA_MODIFIABLE, &modifiable, sizeof (modifiable) },
 		{ CKA_VALUE, value, sizeof (value) },
 		{ CKA_ISSUER, issuer, sizeof (issuer) },
 		{ CKA_SUBJECT, subject, sizeof (subject) },
@@ -461,8 +571,8 @@ check_certificate (CK_SESSION_HANDLE session,
 	};
 
 	/* Note that we don't pass the CKA_INVALID attribute in */
-	rv = test.module->C_GetAttributeValue (session, handle, attrs, 15);
-	assert (rv == CKR_OK);
+	rv = test.module->C_GetAttributeValue (session, handle, attrs, 14);
+	assert_num_eq (rv, CKR_OK);
 
 	/* If this is the cacert3 certificate, check its values */
 	if (memcmp (value, test_cacert3_ca_der, sizeof (test_cacert3_ca_der)) == 0) {
@@ -522,8 +632,6 @@ static void
 test_find_builtin (void)
 {
 	CK_OBJECT_CLASS klass = CKO_NSS_BUILTIN_ROOT_LIST;
-	CK_BBOOL vtrue = CK_TRUE;
-	CK_BBOOL vfalse = CK_FALSE;
 
 	CK_ATTRIBUTE match[] = {
 		{ CKA_CLASS, &klass, sizeof (klass) },
@@ -660,17 +768,18 @@ test_remove_token (void)
 	CK_RV rv;
 
 	rv = test.module->C_OpenSession (test.slots[0], CKF_SERIAL_SESSION, NULL, NULL, &session);
-	assert (rv == CKR_OK);
+	assert_num_eq (rv, CKR_OK);
 
 	rv = test.module->C_FindObjectsInit (session, NULL, 0);
-	assert (rv == CKR_OK);
+	assert_num_eq (rv, CKR_OK);
 
 	rv = test.module->C_FindObjects (session, &handle, 1, &count);
-	assert (rv == CKR_OK);
+	assert_num_eq (rv, CKR_OK);
 	assert_num_eq (1, count);
 
 	rv = test.module->C_DestroyObject (session, handle);
-	assert (rv == CKR_TOKEN_WRITE_PROTECTED);
+	if (rv != CKR_TOKEN_WRITE_PROTECTED)
+		assert_num_eq (rv, CKR_SESSION_READ_ONLY);
 }
 
 static void
@@ -689,17 +798,18 @@ test_setattr_token (void)
 	CK_RV rv;
 
 	rv = test.module->C_OpenSession (test.slots[0], CKF_SERIAL_SESSION, NULL, NULL, &session);
-	assert (rv == CKR_OK);
+	assert_num_eq (rv, CKR_OK);
 
 	rv = test.module->C_FindObjectsInit (session, NULL, 0);
-	assert (rv == CKR_OK);
+	assert_num_eq (rv, CKR_OK);
 
 	rv = test.module->C_FindObjects (session, &handle, 1, &count);
-	assert (rv == CKR_OK);
+	assert_num_eq (rv, CKR_OK);
 	assert_num_eq (1, count);
 
 	rv = test.module->C_SetAttributeValue (session, handle, original, 2);
-	assert (rv == CKR_TOKEN_WRITE_PROTECTED);
+	if (rv != CKR_TOKEN_WRITE_PROTECTED)
+		assert_num_eq (rv, CKR_ATTRIBUTE_READ_ONLY);
 }
 
 static void
@@ -922,6 +1032,148 @@ test_login_logout (void)
 	assert (rv == CKR_USER_NOT_LOGGED_IN);
 }
 
+static void
+test_token_writable (void)
+{
+	CK_TOKEN_INFO info;
+	CK_RV rv;
+
+	rv = test.module->C_GetTokenInfo (test.slots[0], &info);
+
+	assert_num_eq (rv, CKR_OK);
+	assert_num_eq (info.flags & CKF_WRITE_PROTECTED, 0);
+}
+
+static void
+test_session_read_only_create (void)
+{
+	CK_ATTRIBUTE original[] = {
+		{ CKA_CLASS, &data, sizeof (data) },
+		{ CKA_LABEL, "yay", 3 },
+		{ CKA_VALUE, "eight", 5 },
+		{ CKA_TOKEN, &vtrue, sizeof (vtrue) },
+		{ CKA_INVALID }
+	};
+
+	CK_SESSION_HANDLE session;
+	CK_OBJECT_HANDLE handle;
+	CK_RV rv;
+
+	/* Read-only session */
+	rv = test.module->C_OpenSession (test.slots[0], CKF_SERIAL_SESSION,
+	                                 NULL, NULL, &session);
+	assert (rv == CKR_OK);
+
+	/* Create a token object */
+	rv = test.module->C_CreateObject (session, original, 4, &handle);
+	assert_num_eq (rv, CKR_SESSION_READ_ONLY);
+}
+
+static void
+test_create_and_write (void)
+{
+	CK_ATTRIBUTE original[] = {
+		{ CKA_CLASS, &data, sizeof (data) },
+		{ CKA_LABEL, "yay", 3 },
+		{ CKA_VALUE, "eight", 5 },
+		{ CKA_TOKEN, &vtrue, sizeof (vtrue) },
+		{ CKA_INVALID }
+	};
+
+	CK_ATTRIBUTE expected[] = {
+		{ CKA_CLASS, &data, sizeof (data) },
+		{ CKA_LABEL, "yay", 3 },
+		{ CKA_VALUE, "eight", 5 },
+		{ CKA_APPLICATION, "", 0 },
+		{ CKA_OBJECT_ID, "", 0 },
+		{ CKA_INVALID }
+	};
+
+	CK_SESSION_HANDLE session;
+	CK_OBJECT_HANDLE handle;
+	p11_array *parsed;
+	char *path;
+	CK_RV rv;
+	int ret;
+
+	/* Read-only session */
+	rv = test.module->C_OpenSession (test.slots[0], CKF_SERIAL_SESSION | CKF_RW_SESSION,
+	                                 NULL, NULL, &session);
+	assert_num_eq (rv, CKR_OK);
+
+	/* Create a token object */
+	rv = test.module->C_CreateObject (session, original, 4, &handle);
+	assert_num_eq (rv, CKR_OK);
+
+	/* The expected file name */
+	path = p11_path_build (test.directory, "yay.p11-kit", NULL);
+	p11_parser_formats (test.parser, p11_parser_format_persist, NULL);
+	ret = p11_parse_file (test.parser, path, 0);
+	assert_num_eq (ret, P11_PARSE_SUCCESS);
+	free (path);
+
+	parsed = p11_parser_parsed (test.parser);
+	assert_num_eq (parsed->num, 1);
+
+	test_check_attrs (expected, parsed->elem[0]);
+}
+
+static void
+test_modify_and_write (void)
+{
+	CK_ATTRIBUTE original[] = {
+		{ CKA_VALUE, "eight", 5 },
+		{ CKA_CLASS, &data, sizeof (data) },
+		{ CKA_LABEL, "yay", 3 },
+		{ CKA_TOKEN, &vtrue, sizeof (vtrue) },
+		{ CKA_MODIFIABLE, &vtrue, sizeof (vtrue) },
+		{ CKA_INVALID }
+	};
+
+	CK_ATTRIBUTE expected[] = {
+		{ CKA_CLASS, &data, sizeof (data) },
+		{ CKA_LABEL, "yay", 3 },
+		{ CKA_VALUE, "nine", 4 },
+		{ CKA_APPLICATION, "", 0 },
+		{ CKA_OBJECT_ID, "", 0 },
+		{ CKA_INVALID }
+	};
+
+	CK_SESSION_HANDLE session;
+	CK_OBJECT_HANDLE handle;
+	p11_array *parsed;
+	char *path;
+	CK_RV rv;
+	int ret;
+
+	/* Read-only session */
+	rv = test.module->C_OpenSession (test.slots[0], CKF_SERIAL_SESSION | CKF_RW_SESSION,
+	                                 NULL, NULL, &session);
+	assert_num_eq (rv, CKR_OK);
+
+	/* Create a token object */
+	rv = test.module->C_CreateObject (session, original, 5, &handle);
+	assert_num_eq (rv, CKR_OK);
+
+	/* Now modify the object */
+	original[0].pValue = "nine";
+	original[0].ulValueLen = 4;
+
+	rv = test.module->C_SetAttributeValue (session, handle, original, 5);
+	assert_num_eq (rv, CKR_OK);
+
+	/* The expected file name */
+	path = p11_path_build (test.directory, "yay.p11-kit", NULL);
+	ret = p11_parse_file (test.parser, path, 0);
+	assert_num_eq (ret, P11_PARSE_SUCCESS);
+	free (path);
+
+	parsed = p11_parser_parsed (test.parser);
+	assert_num_eq (parsed->num, 1);
+
+	test_check_attrs (expected, parsed->elem[0]);
+}
+
 int
 main (int argc,
       char *argv[])
@@ -933,6 +1185,8 @@ main (int argc,
 	p11_test (test_get_slot_info, "/module/get_slot_info");
 
 	p11_fixture (NULL, NULL);
+	p11_test (test_null_initialize, "/module/initialize-null");
+	p11_test (test_multi_initialize, "/module/initialize-multi");
 	p11_test (test_get_token_info, "/module/get_token_info");
 
 	p11_fixture (setup, teardown);
@@ -953,5 +1207,11 @@ main (int argc,
 	p11_test (test_find_serial_der_mismatch, "/module/find_serial_der_mismatch");
 	p11_test (test_login_logout, "/module/login_logout");
 
+	p11_fixture (setup_writable, teardown);
+	p11_test (test_token_writable, "/module/token-writable");
+	p11_test (test_session_read_only_create, "/module/session-read-only-create");
+	p11_test (test_create_and_write, "/module/create-and-write");
+	p11_test (test_modify_and_write, "/module/modify-and-write");
+
 	return p11_test_run (argc, argv);
 }
