@@ -51,24 +51,25 @@
 
 struct {
 	p11_parser *parser;
+	p11_array *parsed;
 	p11_asn1_cache *cache;
-	p11_index *index;
 } test;
 
 static void
 setup (void *unused)
 {
-	test.index = p11_index_new (NULL, NULL, NULL);
 	test.cache = p11_asn1_cache_new ();
-	test.parser = p11_parser_new (test.index, test.cache);
+	test.parser = p11_parser_new (test.cache);
 	assert_ptr_not_null (test.parser);
+
+	test.parsed = p11_parser_parsed (test.parser);
+	assert_ptr_not_null (test.parsed);
 }
 
 static void
 teardown (void *unused)
 {
 	p11_parser_free (test.parser);
-	p11_index_free (test.index);
 	p11_asn1_cache_free (test.cache);
 	memset (&test, 0, sizeof (test));
 }
@@ -85,12 +86,19 @@ static CK_ATTRIBUTE certificate_match[] = {
 };
 
 static CK_ATTRIBUTE *
-parsed_attrs (CK_ATTRIBUTE *match)
+parsed_attrs (CK_ATTRIBUTE *match,
+              int length)
 {
-	CK_OBJECT_HANDLE handle;
-	handle = p11_index_find (test.index, certificate_match, -1);
-	return p11_index_lookup (test.index, handle);
+	int i;
 
+	if (length < 0)
+		length = p11_attrs_count (match);
+	for (i = 0; i < test.parsed->num; i++) {
+		if (p11_attrs_matchn (test.parsed->elem[i], match, length))
+			return test.parsed->elem[i];
+	}
+
+	return NULL;
 }
 
 static void
@@ -109,14 +117,15 @@ test_parse_der_certificate (void)
 		{ CKA_INVALID },
 	};
 
+	p11_parser_formats (test.parser, p11_parser_format_x509, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der",
 	                      P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
 
 	/* Should have gotten certificate */
-	assert_num_eq (1, p11_index_size (test.index));
+	assert_num_eq (1, test.parsed->num);
 
-	cert = parsed_attrs (certificate_match);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (expected, cert);
 }
 
@@ -136,14 +145,15 @@ test_parse_pem_certificate (void)
 		{ CKA_INVALID },
 	};
 
+	p11_parser_formats (test.parser, p11_parser_format_pem, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.pem",
 	                      P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
 
 	/* Should have gotten certificate  */
-	assert_num_eq (1, p11_index_size (test.index));
+	assert_num_eq (1, test.parsed->num);
 
-	cert = parsed_attrs (certificate_match);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (expected, cert);
 }
 
@@ -157,20 +167,20 @@ test_parse_p11_kit_persist (void)
 		{ CKA_CERTIFICATE_TYPE, &x509, sizeof (x509) },
 		{ CKA_CLASS, &certificate, sizeof (certificate) },
 		{ CKA_VALUE, (void *)verisign_v1_ca, sizeof (verisign_v1_ca) },
-		{ CKA_MODIFIABLE, &falsev, sizeof (falsev) },
 		{ CKA_TRUSTED, &truev, sizeof (truev) },
 		{ CKA_X_DISTRUSTED, &falsev, sizeof (falsev) },
 		{ CKA_INVALID },
 	};
 
+	p11_parser_formats (test.parser, p11_parser_format_persist, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/input/verisign-v1.p11-kit",
 	                      P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
 
 	/* Should have gotten certificate  */
-	assert_num_eq (1, p11_index_size (test.index));
+	assert_num_eq (1, test.parsed->num);
 
-	cert = parsed_attrs (certificate_match);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (expected, cert);
 }
 
@@ -190,16 +200,16 @@ test_parse_openssl_trusted (void)
 	CK_ATTRIBUTE eku_extension[] = {
 		{ CKA_CLASS, &certificate_extension, sizeof (certificate_extension), },
 		{ CKA_OBJECT_ID, (void *)P11_OID_EXTENDED_KEY_USAGE, sizeof (P11_OID_EXTENDED_KEY_USAGE) },
-		{ CKA_X_CRITICAL, &truev, sizeof (truev) },
-		{ CKA_VALUE, "\x30\x0a\x06\x08\x2b\x06\x01\x05\x05\x07\x03\x01", 12 },
+		{ CKA_X_PUBLIC_KEY_INFO, (void *)test_cacert3_ca_public_key, sizeof (test_cacert3_ca_public_key) },
+		{ CKA_VALUE, "\x30\x16\x06\x03\x55\x1d\x25\x01\x01\xff\x04\x0c\x30\x0a\x06\x08\x2b\x06\x01\x05\x05\x07\x03\x01", 24 },
 		{ CKA_INVALID },
 	};
 
 	CK_ATTRIBUTE reject_extension[] = {
 		{ CKA_CLASS, &certificate_extension, sizeof (certificate_extension), },
 		{ CKA_OBJECT_ID, (void *)P11_OID_OPENSSL_REJECT, sizeof (P11_OID_OPENSSL_REJECT) },
-		{ CKA_X_CRITICAL, &falsev, sizeof (falsev) },
-		{ CKA_VALUE, "\x30\x0a\x06\x08\x2b\x06\x01\x05\x05\x07\x03\x04", 12 },
+		{ CKA_X_PUBLIC_KEY_INFO, (void *)test_cacert3_ca_public_key, sizeof (test_cacert3_ca_public_key) },
+		{ CKA_VALUE, "\x30\x1a\x06\x0a\x2b\x06\x01\x04\x01\x99\x77\x06\x0a\x01\x04\x0c\x30\x0a\x06\x08\x2b\x06\x01\x05\x05\x07\x03\x04", 28 },
 		{ CKA_INVALID },
 	};
 
@@ -212,10 +222,10 @@ test_parse_openssl_trusted (void)
 
 	CK_ATTRIBUTE *cert;
 	CK_ATTRIBUTE *object;
-	CK_OBJECT_HANDLE handle;
 	int ret;
 	int i;
 
+	p11_parser_formats (test.parser, p11_parser_format_pem, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3-trusted.pem",
 	                      P11_PARSE_FLAG_ANCHOR);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
@@ -225,18 +235,15 @@ test_parse_openssl_trusted (void)
 	 * - 1 certificate
 	 * - 2 stapled extensions
 	 */
-	assert_num_eq (3, p11_index_size (test.index));
+	assert_num_eq (3, test.parsed->num);
 
 	/* The certificate */
-	cert = parsed_attrs (certificate_match);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (expected[0], cert);
 
 	/* The other objects */
 	for (i = 1; expected[i]; i++) {
-		handle = p11_index_find (test.index, expected[i], 2);
-		assert (handle != 0);
-
-		object = p11_index_lookup (test.index, handle);
+		object = parsed_attrs (expected[i], 2);
 		assert_ptr_not_null (object);
 
 		test_check_attrs (expected[i], object);
@@ -247,6 +254,20 @@ test_parse_openssl_trusted (void)
 static void
 test_parse_openssl_distrusted (void)
 {
+	static const char distrust_public_key[] = {
+		0x30, 0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01,
+		0x05, 0x00, 0x03, 0x81, 0x8d, 0x00, 0x30, 0x81, 0x89, 0x02, 0x81, 0x81, 0x00, 0xdf, 0xc7, 0x0d,
+		0x61, 0xa2, 0x2f, 0xc0, 0x5a, 0xad, 0x45, 0x83, 0x22, 0x33, 0x42, 0xea, 0xec, 0x42, 0x5e, 0xa6,
+		0x0d, 0x42, 0x4c, 0x1c, 0x9a, 0x12, 0x0b, 0x5f, 0xe7, 0x25, 0xf9, 0x8b, 0x83, 0x0c, 0x0a, 0xc5,
+		0x2f, 0x5a, 0x58, 0x56, 0xb8, 0xad, 0x87, 0x6d, 0xbc, 0x80, 0x5d, 0xdd, 0x49, 0x45, 0x39, 0x5f,
+		0xb9, 0x08, 0x3a, 0x63, 0xe4, 0x92, 0x33, 0x61, 0x79, 0x19, 0x1b, 0x9d, 0xab, 0x3a, 0xd5, 0x7f,
+		0xa7, 0x8b, 0x7f, 0x8a, 0x5a, 0xf6, 0xd7, 0xde, 0xaf, 0xa1, 0xe5, 0x53, 0x31, 0x29, 0x7d, 0x9c,
+		0x03, 0x55, 0x3e, 0x47, 0x78, 0xcb, 0xb9, 0x7a, 0x98, 0x8c, 0x5f, 0x8d, 0xda, 0x09, 0x0f, 0xc8,
+		0xfb, 0xf1, 0x7a, 0x80, 0xee, 0x12, 0x77, 0x0a, 0x00, 0x8b, 0x70, 0xfa, 0x62, 0xbf, 0xaf, 0xee,
+		0x0b, 0x58, 0x16, 0xf9, 0x9c, 0x5c, 0xde, 0x93, 0xb8, 0x4f, 0xdf, 0x4d, 0x7b, 0x02, 0x03, 0x01,
+		0x00, 0x01,
+	};
+
 	CK_ATTRIBUTE distrust_cert[] = {
 		{ CKA_CLASS, &certificate, sizeof (certificate), },
 		{ CKA_MODIFIABLE, &falsev, sizeof (falsev) },
@@ -259,16 +280,16 @@ test_parse_openssl_distrusted (void)
 	CK_ATTRIBUTE eku_extension[] = {
 		{ CKA_CLASS, &certificate_extension, sizeof (certificate_extension), },
 		{ CKA_OBJECT_ID, (void *)P11_OID_EXTENDED_KEY_USAGE, sizeof (P11_OID_EXTENDED_KEY_USAGE) },
-		{ CKA_X_CRITICAL, &truev, sizeof (truev) },
-		{ CKA_VALUE, "\x30\x0c\x06\x0a\x2b\x06\x01\x04\x01\x99\x77\x06\x0a\x10", 14 },
+		{ CKA_X_PUBLIC_KEY_INFO, (void *)distrust_public_key, sizeof (distrust_public_key) },
+		{ CKA_VALUE, "\x30\x18\x06\x03\x55\x1d\x25\x01\x01\xff\x04\x0e\x30\x0c\x06\x0a\x2b\x06\x01\x04\x01\x99\x77\x06\x0a\x10", 26 },
 		{ CKA_INVALID },
 	};
 
 	CK_ATTRIBUTE reject_extension[] = {
 		{ CKA_CLASS, &certificate_extension, sizeof (certificate_extension), },
 		{ CKA_OBJECT_ID, (void *)P11_OID_OPENSSL_REJECT, sizeof (P11_OID_OPENSSL_REJECT) },
-		{ CKA_X_CRITICAL, &falsev, sizeof (falsev) },
-		{ CKA_VALUE, "\x30\x0a\x06\x08\x2b\x06\x01\x05\x05\x07\x03\x02", 12 },
+		{ CKA_X_PUBLIC_KEY_INFO, (void *)distrust_public_key, sizeof (distrust_public_key) },
+		{ CKA_VALUE, "\x30\x1a\x06\x0a\x2b\x06\x01\x04\x01\x99\x77\x06\x0a\x01\x04\x0c\x30\x0a\x06\x08\x2b\x06\x01\x05\x05\x07\x03\x02", 28 },
 		{ CKA_INVALID },
 	};
 
@@ -281,7 +302,6 @@ test_parse_openssl_distrusted (void)
 
 	CK_ATTRIBUTE *cert;
 	CK_ATTRIBUTE *object;
-	CK_OBJECT_HANDLE handle;
 	int ret;
 	int i;
 
@@ -289,6 +309,7 @@ test_parse_openssl_distrusted (void)
 	 * OpenSSL style is to litter the blacklist in with the anchors,
 	 * so we parse this as an anchor, but expect it to be blacklisted
 	 */
+	p11_parser_formats (test.parser, p11_parser_format_pem, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/distrusted.pem",
 	                      P11_PARSE_FLAG_ANCHOR);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
@@ -298,16 +319,13 @@ test_parse_openssl_distrusted (void)
 	 * - 1 certificate
 	 * - 2 stapled extensions
 	 */
-	assert_num_eq (3, p11_index_size (test.index));
-	cert = parsed_attrs (certificate_match);
+	assert_num_eq (3, test.parsed->num);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (expected[0], cert);
 
 	/* The other objects */
 	for (i = 1; expected[i]; i++) {
-		handle = p11_index_find (test.index, expected[i], 2);
-		assert (handle != 0);
-
-		object = p11_index_lookup (test.index, handle);
+		object = parsed_attrs (expected[i], 2);
 		assert_ptr_not_null (object);
 
 		test_check_attrs (expected[i], object);
@@ -331,6 +349,7 @@ test_parse_anchor (void)
 	CK_ATTRIBUTE *cert;
 	int ret;
 
+	p11_parser_formats (test.parser, p11_parser_format_x509, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der",
 	                      P11_PARSE_FLAG_ANCHOR);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
@@ -339,9 +358,9 @@ test_parse_anchor (void)
 	 * Should have gotten:
 	 * - 1 certificate
 	 */
-	assert_num_eq (1, p11_index_size (test.index));
+	assert_num_eq (1, test.parsed->num);
 
-	cert = parsed_attrs (certificate_match);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (cacert3, cert);
 }
 
@@ -360,14 +379,15 @@ test_parse_thawte (void)
 		{ CKA_INVALID },
 	};
 
+	p11_parser_formats (test.parser, p11_parser_format_pem, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/thawte.pem",
 	                      P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
 
 	/* Should have gotten certificate  */
-	assert_num_eq (1, p11_index_size (test.index));
+	assert_num_eq (1, test.parsed->num);
 
-	cert = parsed_attrs (certificate_match);
+	cert = parsed_attrs (certificate_match, -1);
 	test_check_attrs (expected, cert);
 }
 
@@ -380,6 +400,7 @@ test_parse_invalid_file (void)
 
 	p11_message_quiet ();
 
+	p11_parser_formats (test.parser, p11_parser_format_x509, NULL);
 	ret = p11_parse_file (test.parser, "/nonexistant",
 	                      P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_FAILURE, ret);
@@ -394,6 +415,7 @@ test_parse_unrecognized (void)
 
 	p11_message_quiet ();
 
+	p11_parser_formats (test.parser, p11_parser_format_x509, NULL);
 	ret = p11_parse_file (test.parser, SRCDIR "/files/unrecognized-file.txt",
 	                      P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_UNRECOGNIZED, ret);
@@ -402,121 +424,22 @@ test_parse_unrecognized (void)
 }
 
 static void
-test_duplicate (void)
+test_parse_no_asn1_cache (void)
 {
-	CK_ATTRIBUTE cacert3[] = {
-		{ CKA_CLASS, &certificate, sizeof (certificate) },
-		{ CKA_CERTIFICATE_TYPE, &x509, sizeof (x509) },
-		{ CKA_VALUE, (void *)test_cacert3_ca_der, sizeof (test_cacert3_ca_der) },
-		{ CKA_MODIFIABLE, &falsev, sizeof (falsev) },
-		{ CKA_TRUSTED, &falsev, sizeof (falsev) },
-		{ CKA_X_DISTRUSTED, &falsev, sizeof (falsev) },
-		{ CKA_INVALID },
-	};
-
-	CK_OBJECT_HANDLE *handles;
-	CK_ATTRIBUTE *cert;
+	p11_parser *parser;
 	int ret;
 
-	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der", 0);
+	parser = p11_parser_new (NULL);
+	assert_ptr_not_null (parser);
+
+	p11_parser_formats (parser, p11_parser_format_x509, NULL);
+	ret = p11_parse_file (parser, SRCDIR "/files/cacert3.der", P11_PARSE_FLAG_NONE);
 	assert_num_eq (P11_PARSE_SUCCESS, ret);
 
-	p11_message_quiet ();
+	/* Should have gotten certificate  */
+	assert_num_eq (1, p11_parser_parsed (parser)->num);
 
-	/* This shouldn't be added, should print a message */
-	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der", 0);
-	assert_num_eq (P11_PARSE_SUCCESS, ret);
-
-	assert (strstr (p11_message_last (), "duplicate") != NULL);
-
-	p11_message_loud ();
-
-	/* Should only be one certificate since the above two are identical */
-	handles = p11_index_find_all (test.index, cacert3, 2);
-	assert_ptr_not_null (handles);
-	assert (handles[0] != 0);
-	assert (handles[1] == 0);
-
-	cert = p11_index_lookup (test.index, handles[0]);
-	test_check_attrs (cacert3, cert);
-
-	free (handles);
-}
-
-static void
-test_duplicate_priority (void)
-{
-	CK_ATTRIBUTE cacert3[] = {
-		{ CKA_CLASS, &certificate, sizeof (certificate) },
-		{ CKA_VALUE, (void *)test_cacert3_ca_der, sizeof (test_cacert3_ca_der) },
-		{ CKA_CERTIFICATE_TYPE, &x509, sizeof (x509) },
-		{ CKA_MODIFIABLE, &falsev, sizeof (falsev) },
-		{ CKA_INVALID },
-	};
-
-	CK_ATTRIBUTE trusted[] = {
-		{ CKA_CLASS, &certificate, sizeof (certificate) },
-		{ CKA_VALUE, (void *)test_cacert3_ca_der, sizeof (test_cacert3_ca_der) },
-		{ CKA_CERTIFICATE_TYPE, &x509, sizeof (x509) },
-		{ CKA_TRUSTED, &truev, sizeof (truev) },
-		{ CKA_X_DISTRUSTED, &falsev, sizeof (falsev) },
-		{ CKA_INVALID },
-	};
-
-	CK_ATTRIBUTE distrust[] = {
-		{ CKA_CLASS, &certificate, sizeof (certificate) },
-		{ CKA_VALUE, (void *)test_cacert3_ca_der, sizeof (test_cacert3_ca_der) },
-		{ CKA_CERTIFICATE_TYPE, &x509, sizeof (x509) },
-		{ CKA_TRUSTED, &falsev, sizeof (falsev) },
-		{ CKA_X_DISTRUSTED, &truev, sizeof (truev) },
-		{ CKA_INVALID },
-	};
-
-	CK_OBJECT_HANDLE *handles;
-	CK_ATTRIBUTE *cert;
-	int ret;
-
-	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der", 0);
-	assert_num_eq (P11_PARSE_SUCCESS, ret);
-
-	p11_message_quiet ();
-
-	/* This shouldn't be added, should print a message */
-	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der",
-	                      P11_PARSE_FLAG_ANCHOR);
-	assert_num_eq (P11_PARSE_SUCCESS, ret);
-
-	assert (strstr (p11_message_last (), "duplicate") != NULL);
-
-	p11_message_loud ();
-
-	/* We should now find the trusted certificate */
-	handles = p11_index_find_all (test.index, cacert3, 2);
-	assert_ptr_not_null (handles);
-	assert (handles[0] != 0);
-	assert (handles[1] == 0);
-	cert = p11_index_lookup (test.index, handles[0]);
-	test_check_attrs (trusted, cert);
-	free (handles);
-
-	/* Now add a distrutsed one, this should override the trusted */
-
-	p11_message_quiet ();
-
-	ret = p11_parse_file (test.parser, SRCDIR "/files/cacert3.der",
-	                      P11_PARSE_FLAG_BLACKLIST);
-	assert_num_eq (P11_PARSE_SUCCESS, ret);
-
-	p11_message_loud ();
-
-	/* We should now find the distrusted certificate */
-	handles = p11_index_find_all (test.index, cacert3, 2);
-	assert_ptr_not_null (handles);
-	assert (handles[0] != 0);
-	assert (handles[1] == 0);
-	cert = p11_index_lookup (test.index, handles[0]);
-	test_check_attrs (distrust, cert);
-	free (handles);
+	p11_parser_free (parser);
 }
 
 int
@@ -533,7 +456,9 @@ main (int argc,
 	p11_test (test_parse_thawte, "/parser/parse_thawte");
 	p11_test (test_parse_invalid_file, "/parser/parse_invalid_file");
 	p11_test (test_parse_unrecognized, "/parser/parse_unrecognized");
-	p11_test (test_duplicate, "/parser/duplicate");
-	p11_test (test_duplicate_priority, "/parser/duplicate_priority");
+
+	p11_fixture (NULL, NULL);
+	p11_test (test_parse_no_asn1_cache, "/parser/null-asn1-cache");
+
 	return p11_test_run (argc, argv);
 }
